require 'freshbooks.rb'
require 'dotenv'
require 'terminal-table'


class Task
  def self.find(id)
    memoize :task, id do
      FreshBooks::Task.get id
    end
  end

  def self.memoize(*keys)
    @_cache ||= {}
    @_cache[keys.join('-')] ||= yield
  end
end

class Projects
  Project = Struct.new(:number, :id, :name)

  def self.pick_project
    new.select_project
  end

  def select_project
    puts project_table
    print "\nChoose a project number: "
    project = ask_project_select
    FreshBooks::Project.get project.id if project
  end

  private

  def project_table
    Terminal::Table.new do |t|
      t << ['Number', 'Project']
      t << :separator

      puts "Loading projects..."
      projects.each do |project|
        t << [project.number, project.name]
      end
    end
  end

  def ask_project_select
    project_number = gets.chomp
    project = find_project_by_number project_number.to_i
    puts "Project not found" if project.nil?
    project
  end

  def projects
    @_projects ||= FreshBooks::Project.list.map { |project|
      Project.new next_number, project.project_id, project.name
    }
  end

  def find_project_by_number(number)
    projects.find { |project| project.number == number }
  end

  def next_number
    @n ||= 0
    @n += 1 # always increment. This way we'll start with 1
  end
end


class TimeEntries
  def self.render_project(project)
    new(project).render
  end

  def initialize(project)
    @project = project
    @total_hours = 0
  end

  def render
    puts time_entries_table
  end

  private

  attr_reader :project

  def time_entries_table
    @_time_entries_table ||= Terminal::Table.new title: title do |t|
      add_header t
      t.add_separator
      add_entries t
      t.add_separator
      add_totals t
    end
  end

  def title
    project.name
  end

  def add_header(t)
    t << ['Task', 'Date', 'Hours', 'Notes']
  end

  def add_entries(t)
    entries.each do |entry|
      task = Task.find entry.task_id
      @total_hours += entry.hours
      t << [task.name, entry.date.inspect, render_hours(entry.hours), entry.notes]
    end
  end

  def add_totals(t)
    t << [{ value: 'Project Budget', colspan: 2 }, '%7.2f' % project_budget, '']
    t << [{ value: 'Total Hours', colspan: 2 }, '%7.2f' % @total_hours, '']

    if !project_budget.zero?
      remaining = project_budget - @total_hours
      t << [{ value: 'Time Remaining', colspan: 2 }, '%7.2f' % remaining, '']
    end
  end

  def render_hours(hours)
    '%7.2f' % hours
  end

  def project_budget
    @_project_budget ||= project.budget.hours.to_f
  end

  def entries
    puts "\nLoading...\n\n"
    FreshBooks::TimeEntry.list project_id: project_id
  end

  def project_id
    project.project_id
  end
end


module Freshbookstimestats
  def self.load(dir = nil)
    Dotenv.load(*Array(dir))
    FreshBooks::Base.establish_connection(ENV['FRESHBOOKS_DOMAIN'], ENV['FRESHBOOKS_AUTH_TOKEN'])
  end

  def self.run
    project = Projects.pick_project
    TimeEntries.render_project project if project
  end
end


